import numpy as np
import pandas as pd
from pandas.core.base import PandasObject


## Efficient Rolling Window Calculations

def numpy_rolling_window(data, window):
    shape = data.shape[:-1] + (data.shape[-1] - window + 1, window)
    strides = data.strides + (data.strides[-1],)
    return np.lib.stride_tricks.as_strided(data, shape=shape, strides=strides)


def numpy_rolling_series(func):
    def func_wrapper(data, window, as_source=False):
        series = data.values if isinstance(data, pd.Series) else data

        new_series = np.empty(len(series)) * np.nan
        calculated = func(series, window)
        new_series[-len(calculated):] = calculated

        if as_source and isinstance(data, pd.Series):
            return pd.Series(index=data.index, data=new_series)

        return new_series
    return func_wrapper

@numpy_rolling_series
def numpy_rolling_mean(data, window, as_source=False):
    return np.mean(numpy_rolling_window(data, window), axis=-1)

@numpy_rolling_series
def numpy_rolling_std(data, window, as_source=False):
    return np.std(numpy_rolling_window(data, window), axis=-1, ddof=1)


## Indicator Functions

def true_range(bars):
    return pd.DataFrame({
        "h_l": bars['high'] - bars['low'],
        "h_c": abs(bars['high'] - bars['close'].shift(1)),
        "l_c": abs(bars['low'] - bars['close'].shift(1))
    }).max(axis=1)


def atr(bars, window=14):
    tr = true_range(bars)
    res = numpy_rolling_mean(tr, window)
    return pd.Series(res)


def rsi(series, window=14):
    deltas = np.diff(series)
    seed = deltas[:window + 1]

    ups = seed[seed > 0].sum() / window
    downs = -seed[seed < 0].sum() / window
    rsival = np.zeros_like(series)
    rsival[:window] = 100. - 100. / (1. + ups / downs)

    for i in range(window, len(series)):
        delta = deltas[i - 1]
        if delta > 0:
            upval = delta
            downval = 0
        else:
            upval = 0
            downval = -delta

        ups = (ups * (window - 1) + upval) / window
        downs = (downs * (window - 1.) + downval) / window
        rsival[i] = 100. - 100. / (1. + ups / downs)

    return pd.Series(index=series.index, data=rsival)


def bollinger_bands(series, window=20, stds=2):
    ma = numpy_rolling_mean(series, window=window)
    std = numpy_rolling_std(series, window=window)
    upper = ma + std * stds
    lower = ma - std * stds

    return pd.DataFrame(index=series.index, data={
        'upper': upper,
        'mid': ma,
        'lower': lower
    })


def zscore(series, window=20, stds=1):
    std = numpy_rolling_std(series, window)
    mean = numpy_rolling_mean(series, window)
    return (series - mean) / (std * stds)

## Extend PandasObject
PandasObject.true_range = true_range
PandasObject.atr = atr
PandasObject.rsi = rsi
PandasObject.bollinger_bands = bollinger_bands
PandasObject.zscore = zscore


## Usage
import lakeapi
import matplotlib.pyplot as plt

# Get freesample data
lakeapi.use_sample_data(anonymous_access = True)
data = lakeapi.load_data(table = 'trades', exchanges = ['BINANCE'], symbols = ['BTC-USDT']).set_index('origin_time')
data = data.iloc[-1000:]

# Calculate indicator
bands = data['price'].bollinger_bands(window=50)

# Plot results
plt.figure(figsize=(12, 8))
plt.plot(data['price'], label='Price')
plt.plot(bands['upper'], label='Upper Band', linestyle='--')
plt.plot(bands['mid'], label='Moving Avg')
plt.plot(bands['lower'], label='Lower Band', linestyle='--')
plt.gca().get_xaxis().set_visible(False)
plt.legend()
plt.savefig('pandas_trading_indicators.png')
